import sys
import os.path
import operator

import cv2
import numpy as np

#Contours and morphology reference
#http://docs.opencv.org/doc/tutorials/imgproc/opening_closing_hats/opening_closing_hats.html
#http://docs.opencv.org/doc/tutorials/imgproc/erosion_dilatation/erosion_dilatation.html
#http://opencvpython.blogspot.co.at/2012/06/hi-this-article-is-tutorial-which-try.html
#https://opencv-python-tutroals.readthedocs.org/
#
#not really relevant, but other explanations of morphology
#http://www.aishack.in/2010/07/mathematical-morphology/
#http://www.aishack.in/2010/07/mathematical-morphology-%E2%80%93-composite-operations/
#http://www.aishack.in/2010/07/mathematical-morphology-in-opencv/

#Thoughts on heuristic based fly head detection
#
#afaict, none of the image moments from the contour can tell you the 'fat' or thin end
#of the contour.
#http://opencvpython.blogspot.co.at/2012/04/contour-features.html
#Maybe could try shape matching... but
#
#we can't match a fly from the contour alone (because matchshapes only uses hu moments
#to return similarity, but maybe other simple operations can elucidate structure. skeletizing
#the fly might show asymetry
#http://felix.abecassis.me/2011/09/opencv-morphological-skeleton/

class CPUTemplateMatcher:

    res = None

    def __init__(self, method=cv2.TM_SQDIFF_NORMED):
        self._tmpl = None
        self._tmplflip = None
        self._method = method

    def load_template(self, tmpl):
        self.set_template(cv2.imread(tmpl, cv2.cv.CV_LOAD_IMAGE_GRAYSCALE))

    def set_template(self, tmpl):
        self._tmpl = tmpl
        self._tmplflip = cv2.flip(self._tmpl, 0)

    def _parse_res(self, _res):
        _min_val, _max_val, _min_loc, _max_loc = cv2.minMaxLoc(_res)
        if self._method == cv2.TM_SQDIFF_NORMED or self._method == cv2.TM_SQDIFF:
            loc = _min_loc
            val = _min_val
            cmp_func = operator.le
        else:
            loc = _max_loc
            val = _max_val
            cmp_func = operator.ge
        return loc,val,cmp_func

    def _check_res(self, val):
        if self._method == cv2.TM_SQDIFF_NORMED:
            #good squared differences are < 0.4
            return val < 0.9
        return True

    def match_template(self, img):
        #do the upright template first
        self.res = cv2.matchTemplate(img, self._tmpl, self._method)
        uloc,uval,_ = self._parse_res(self.res)
        return uloc,uval,self._check_res(uval)

    def match_template_both_ways(self, img):
        #do the upright template first
        ures = cv2.matchTemplate(img, self._tmpl, self._method)
        uloc,uval,ucmp = self._parse_res(ures)

        #do the flipped template
        uresflip = cv2.matchTemplate(img, self._tmplflip, self._method)
        floc,fval,_ = self._parse_res(uresflip)

        if ucmp(uval,fval):
            flipped = False
            loc = uloc
            val = uval
            self.res = ures
        else:
            flipped = True
            loc = floc
            val = fval
            self.res = uresflip

        return loc,val,self._check_res(val),flipped


class FlyBodyMatch(object):

    def __init__(self, vis_callback=None, downsample=1, templateimg=None, templatecontour=None, mincontourarea=2000, check_flipped=False, use_gpu=False):
        self._N = downsample
        self._min_contour_area = mincontourarea
        self._vis_callback = None
        self._check_flipped = check_flipped

        self.threshold = -1
        self.should_visualize = False
        self.mid_adjust_y = 0.5
        #trigger the setter below
        self.smooth = 3

        if (templateimg is not None) and os.path.isfile(templateimg):
            self._template = cv2.imread(templateimg, cv2.cv.CV_LOAD_IMAGE_GRAYSCALE)
            self._template_contour_area,_ = self.calculate_largest_contour(self._template)
        elif templatecontour is not None:
            raise Exception("Not implemented")
        else:
            self._template = None
            self._template_contour_area = 0

        if self._template is not None:
            self._template_w, self._template_h = self._template.shape[::-1]

        if vis_callback is not None:
            self.enable_visualization(vis_callback)

        self._use_gpu = False
        klass = CPUTemplateMatcher
        try:
            if use_gpu:
                import cvtempl
                if cvtempl.is_available():
                    klass = cvtempl.GPUTemplateMatcher
                    self._use_gpu = True
        except ImportError:
            pass

        self._matcher = klass(method=cv2.TM_SQDIFF_NORMED)
        self._matcher.set_template(self._template)

    @property
    def gpu_enabled(self):
        return self._use_gpu

    @property
    def downsample_factor(self):
        return self._N if self._N else 1

    @property
    def smooth(self):
        return self._smooth

    @smooth.setter
    def smooth(self, v):
        self._smooth = v
        self._smooth_kernel = np.ones((self._smooth, self._smooth), np.uint8)

    def get_template_contour_area(self):
        return self._template_contour_area

    def visualize(self, name, arr):
        if self.should_visualize:
            self._vis_callback(name, arr)

    def enable_visualization(self, vis_callback):
        self.should_visualize = True
        self._vis_callback = vis_callback

    def calculate_largest_contour(self, img):
        """
        return the area of the largest contour in the image
        """
        #RETR_EXTERNAL returns only external, likely largest, contours
        #CHAIN_APPROX_SIMPLE connects neighabouring points a little

        contours0, hierarchy = cv2.findContours(img.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        cnt = None

        #keep the largest contour above self._min_contour_area
        cnt_size = 0
        for _cnt in contours0:
            _cnt_size = cv2.contourArea(_cnt)
            if _cnt_size > cnt_size:
                cnt = _cnt
                cnt_size = _cnt_size
        return cnt_size, cnt

    def get_fly_contour(self, iimg):
        """
        Extract and return an upright contour of the fly from the input image.
        The following algorithm is used
            1. downsample
            2. threshold to binary (fly = white)
            3. filter using morphology open operator
            4. extract the largest contour
            5. fit an ellipse to the largest contour and get the
               affine transform matrix (from the ellipse parameters)
               required to right the fly (head at top or bottom)
            6. transform the largest contour according the previous step
               and center it about some fixed point.

        This function also returns some additional information so additional
        image processing steps can be performed.

        The contour is centered on 0,0. The contour is an array of size (npts,2).
        To use OpenCV's contour functions it must be resized to (npts,1,2)...

        @returns: (downsampled_image, img_used_for_contour_detection,
                   ellipse, contour, rotation_matrix)
        """
        assert iimg.ndim == 2

        #if indexing looks switched here remember that primarily this is an
        #opencv program and therefor uses image coordinates
        #img = cv2.someFunction(x,y)
        #
        #numpy uses indexing which is equivilently
        #val = img[y,x]

        #downsample by N
        if self._N:
            N = self._N
            img = iimg[::self._N,::self._N]
        else:
            N = 1
            img = iimg

        #threshold and invert to make fly white (needed for morphological operations)
        thresh, img2 = cv2.threshold(img, self.threshold, 255,
                                    (cv2.THRESH_BINARY_INV | cv2.THRESH_OTSU)
                                    if self.threshold == -1 else cv2.THRESH_BINARY_INV)

        #morph open to smooth noise. mophilogical operations are faster than filtering
        if self.smooth:
            #the morph operation does a good enough job of filling the middle of the fly
            #that the combination of RETR_EXTERNAL and taking the largest contour means
            #the fly does not have to be filled specifically with cv2.floodFill just to get
            #a robust contour
            img3 = cv2.morphologyEx(img2, cv2.MORPH_OPEN, self._smooth_kernel)
        else:
            img3 = img2

        imgcontour_h, imgcontour_w = img3.shape[0], img3.shape[1]

        cnt_size, cnt = self.calculate_largest_contour(img3)
        if cnt_size < self._min_contour_area:
            cnt_size = 0
            cnt = None

        ellipse = None
        #need 5 points to fit an ellipse
        if (cnt is not None) and (len(cnt) > 5):
            ellipse = cv2.fitEllipse(cnt)

            if self.should_visualize:
                imgc = cv2.cvtColor(img3, cv2.COLOR_GRAY2BGR)
                #draw the ellipse
                cv2.ellipse(imgc,ellipse,(0,255,0))
                (x,y),(w,h),theta = ellipse
                #draw the ellipse center
                cv2.circle(imgc, (int(x),int(y)),3,(0,255,0),2)
                #draw the fly contour
                cv2.drawContours(imgc, [cnt], 0, (255,0,0))
                self.visualize('contour', imgc)

        #no fly detected
        if not ellipse:
            self.visualize('contour', img3)
            return img, img3, None, None, None, None, {}

        #calculate the rotation matrix required rotate the ellipse so it
        #stands vertically
        (x,y), (w,h), theta = ellipse


        rm = cv2.getRotationMatrix2D((x, y), theta, 1.0)
        rm_origin = np.array((x, y))

        if 0:
            #not a useful debug image because I do not center it like I do the
            #contour
            img5 = cv2.warpAffine(img2, rm,
                                  (500, 500),
                                  borderMode=cv2.BORDER_CONSTANT,
                                  borderValue=0)
            self.visualize('rotated_image', img5)

        ncntpoints = len(cnt)
        #remove the extra dimension (n,1,2)
        _cnt = cnt.reshape((ncntpoints, 2))
        #make homogenous coordinates
        _cnth = np.c_[_cnt, np.ones(ncntpoints)]
        #affine transform to be right way up
        uprightcnt = rm.dot(_cnth.T).T

        extra = {'contour_area':cnt_size, 'threshold':thresh}
        return img, img3, ellipse, uprightcnt, rm, rm_origin, extra

    def detect_fly_body(self, iimg):
        """
        Detect fly (no front/back information)

        @returns: the center of the ellipse around the body

        Performs the contour-finding operation.
        """
        x, y = None, None
        img, imgcontour, ellipse, contour, rm, rm_origin, extra = self.get_fly_contour(iimg)
        if ellipse is not None:
            (_x, _y), (w, h), theta = ellipse
            #scale up to original image size
            x, y = _x*self._N, _y*self._N
        return x, y, extra

    def _detect_fly_head_method_template(self, img, imgcontour, ellipse, contour, rm, rm_origin):
        #center the contour on a portrait orientation image
        ds_img_h, ds_img_w = imgcontour.shape
        c_img_h, c_img_w = ds_img_w, ds_img_h
        img_filled_contour = np.zeros((c_img_h, c_img_w), dtype=np.uint8)

        #the contour has been rotated upright about rm_origin
        #move the contour to the middle of the portrait image
        contour = contour - rm_origin + np.array((c_img_w/2, c_img_h/2))
        #add the extra dimension required by opencv drawContours
        contour = contour.reshape((len(contour), 1, 2))
        #fill the contour
        cv2.drawContours(img_filled_contour, [contour.astype(np.int32)], 0, (255, 255, 255), -1)

        if self._template is None:
            #for debugging, save the template
            self.visualize('upright', img_filled_contour)
            cv2.imwrite('frame.png', img_filled_contour)
            return (None,None),0
        else:
            self.visualize('template', self._template)

        if self._check_flipped:
            top_left,match_val,val_ok,match_flipped = self._matcher.match_template_both_ways(img_filled_contour)
        else:
            top_left,match_val,val_ok = self._matcher.match_template(img_filled_contour)
            match_flipped = False

        if not val_ok:
            return (None,None),0

        bottom_right = (top_left[0] + self._template_w, top_left[1] + self._template_h)
        #assume the head is in the middle of the template...
        head_x = top_left[0] + self._template_w/2
        if match_flipped:
            head_y = top_left[1] + (self._template_h * self.mid_adjust_y)
        else:
            head_y = top_left[1] + (self._template_h * (1.0-self.mid_adjust_y))

        if self.should_visualize:
            imgvis = img_filled_contour.copy()
            cv2.rectangle(imgvis, top_left, bottom_right, 255, 2)
            cv2.circle(imgvis, (int(head_x), int(head_y)), 5,(0,0,0),2)
            self.visualize('match', imgvis)

            if self._matcher.res is not None:
                res = self._matcher.res.copy()
                #normalize the result to make it look better
                cv2.normalize(res, res, 0, 255, cv2.NORM_MINMAX, -1)
                self.visualize('result', res.astype(np.uint8))

        #remove the offsets added above
        head = np.array((head_x, head_y)) + rm_origin - np.array((c_img_w/2, c_img_h/2))
        #and make homogenous coords
        headh = np.r_[head, 1]

        #reverse the transform to return the coordinates in the original image
        rmi = cv2.invertAffineTransform(rm)
        #and account for scale factor
        head = rmi.dot(headh) * self.downsample_factor

        return head, 1.0

    def detect_fly_head(self, iimg):
        """
        Detect fly with front/back information

        Performs the contour-finding operation and then template
        matching to detect which side is the head.

        """
        x, y, ex, ey = None, None, None, None

        img, imgcontour, ellipse, contour, rm, rm_origin, extra = self.get_fly_contour(iimg)

        if ellipse is not None:
            #always calculate the COM of the ellipse
            (_ex, _ey), (w, h), theta = ellipse
            ex, ey = _ex*self._N, _ey*self._N

            (x, y), conf = self._detect_fly_head_method_template(img, imgcontour, ellipse, contour, rm, rm_origin)

        return x, y, ex, ey, extra

if __name__ == "__main__":
    import os
    import sys
    import time
    import cv2
    from flyio import FMFHeadMatcher

    VIS = 'SHOW' in os.environ

    _viswindows = {}
    def _cv2_vis_cb(name, img):
        if name not in _viswindows:
            cv2.namedWindow(name)
        cv2.imshow(name, img)

    fmf = str(sys.argv[1])
    hm = FMFHeadMatcher(fmf)

    blob = FlyBodyMatch(downsample=4)
    if VIS:
        cv2.startWindowThread()
        blob.enable_visualization(_cv2_vis_cb)

    t0, n = time.time(), 0
    while 1:
        img = hm.get_next_frame()
        if img is None:
            break

        n += 1
        blob.detect_fly_head(img)

    t1 = time.time()
    print "%.1f ms per frame" % ((t1-t0)/n * 1000)



