from __future__ import with_statement, division

import warnings
import threading
import time
import collections
import os.path
import pkg_resources

import cv2
import numpy as np

from traitsui.api import EnumEditor

from motmot.fview.traited_plugin import *

import blob

CV2_VERSION_REQUIRED = '2.4.6.1'

class FviewHeadTrack(HasTraits_FViewPlugin):
    plugin_name = 'Through-The-Mirrors (TTM) Tracker'

    enabled = traits.Bool(False)
    enable_which = traits.Enum('head', 'body', cols=2, mode='list')

    visualize = traits.Bool(True)

    target_x = traits.Int(0)
    target_x_max = traits.Int(5000)
    target_y = traits.Int(0)
    target_y_max = traits.Int(5000)

    head_offset_y = traits.Float(0.5)

    threshold = traits.Range(-1, 255, -1, is_float=False, mode='slider')
    smooth = traits.Range(0, 10, 3, is_float=False, mode='spinner')

    traits_view = View(Group(
                        Group(
                            Item(name='enabled'),
                            Item(name='enable_which',
                                 editor=EnumEditor(values={
        'head' : 'centroid then template matching',
        'body' : 'centroid only',
        }),
                                 ),
                        ),
                        Group(
                            Item(name='visualize'),
                        ),
                        Group(
                            Item(name='target_x', label='Laser center X',
                                 editor=RangeEditor(
                                                         low = 0,
                                                         high_name = 'target_x_max',
                                                         mode = 'slider',
                                                         is_float = False)),
                            Item(name='target_y', label='Laser center Y',
                                 editor=RangeEditor(
                                                         low = 0,
                                                         high_name = 'target_y_max',
                                                         mode = 'slider',
                                                         is_float = False)),
                            Item(name='head_offset_y', label='Template fraction:',
                                 editor=RangeEditor(
                                                         low = 0.0,
                                                         high = 1.0,
                                                         mode = 'slider',
                                                         is_float = True)),
                            label = 'Targeting parameters',
                            show_border = True,
                        ),
                        Group( Item(name='threshold'),
                               Item(name='smooth'),
                               label = 'Image processing parameters',
                               show_border = True,
                               ),
                        ),
                       )

    def __init__(self,wx_parent,fview_options):
        super(FviewHeadTrack,self).__init__(wx_parent)

        self._wx_parent = wx_parent

        if cv2.__version__ != CV2_VERSION_REQUIRED:
            raise Exception("Opencv version %s required" % CV2_VERSION_REQUIRED)

        self.have_ros = False
        if fview_options.get('have_ros'):
            import roslib.packages
            try:
                roslib.load_manifest('flymad')
                import rospy
                import flymad.msg
                from std_msgs.msg import Float32

                self._pub = rospy.Publisher('/flymad/laser_head_delta',
                                            flymad.msg.HeadDetect,
                                            tcp_nodelay=True)
                self._no_detect = flymad.msg.HeadDetect.NO_DETECT
                self._target_map = {"head":flymad.msg.HeadDetect.TARGET_HEAD,
                                    "body":flymad.msg.HeadDetect.TARGET_BODY}
                blobopts = {
                    "downsample":int(rospy.get_param("/ttm/headtrack_downsample",2)),
                    "templateimg":rospy.get_param("/ttm/headtrack_templateimg","head4.png"),
                    "mincontourarea":int(rospy.get_param("/ttm/headtrack_mincontourarea",3000)),
                    "check_flipped":bool(int(rospy.get_param("/ttm/headtrack_checkflipped",1))),
                    "use_gpu":bool(int(rospy.get_param("/ttm/headtrack_usegpu",0))),
                }

                for k,v in blobopts.items():
                    rospy.loginfo("head tracker config: %s=%r" % (k,v))

                self.have_ros = True
            except roslib.packages.InvalidROSPkgException:
                pass

        if not self.have_ros:
            self._pub = None
            self._no_detect = 1e6
            self._target_map = {"head":1,"body":2}

            blobopts = {
                "downsample":3,
                "templateimg":"fly2.png",
                "mincontourarea":2000,
                "check_flipped":True,
                "use_gpu":False,
            }

            for k,v in blobopts.items():
                print "head tracker config: %s=%r" % (k,v)

        if blobopts["templateimg"] and not os.path.isabs(blobopts["templateimg"]):
            blobopts["templateimg"] = pkg_resources.resource_filename(
                    __name__,
                    os.path.join("data",blobopts["templateimg"])
            )

        self._viswindows = {}
        self._visimgs = {}
        self._vislock = threading.Lock()

        tid = wx.NewId()
        self._vis_timer = wx.Timer(self._wx_parent, tid)
        wx.EVT_TIMER(self._wx_parent, tid, self._update_wx_visualizations)
        self._vis_timer.Start(200) # 5 times per sec

        self.blob = blob.FlyBodyMatch(vis_callback=self._visualize_cb,
                                      **blobopts)
        template_contour_area = self.blob.get_template_contour_area()

        if self.have_ros:
            rospy.loginfo("head tracker GPU initialized OK: %s" % self.blob.gpu_enabled)
            tmp_pub = rospy.Publisher('/flymad/template_contour_area',
                                      Float32,
                                      latch=True)
            tmp_pub.publish( Float32( template_contour_area ) )
        else:
            print "head tracker GPU initialized OK: %s" % self.blob.gpu_enabled

        #only show the target for a few seconds after adjusting it
        self._show_target_time = time.time()

    def _update_wx_visualizations(self, evt):
        with self._vislock:
            for name, arr in self._visimgs.items():
                if name not in self._viswindows:
                    f = wx.Frame(None,title=name,size=(arr.shape[1],arr.shape[0]))
                    w = wxvideo.DynamicImageCanvas(f)
                    self._viswindows[name] = w
                    f.Show()

                self._viswindows[name].update_image_and_drawings(
                    'camera', arr, format='MONO8' if arr.ndim == 2 else 'RGB8',
                )
                #no idea what eraseBackground does. it works with True and False
                self._viswindows[name].Refresh(eraseBackground=False)

    def _visualize_cb(self, name, arr):
        with self._vislock:
            self._visimgs[name] = arr.copy()

    def _threshold_fired(self):
        self.blob.threshold = self.threshold

    def _visualize_fired(self):
        self.blob.should_visualize = self.visualize

    def _smooth_fired(self):
        self.blob.smooth = self.smooth

    def _target_x_fired(self):
        self._show_target_time = time.time()

    def _target_y_fired(self):
        self._show_target_time = time.time()

    def _head_offset_y_fired(self):
        self.blob.mid_adjust_y = self.head_offset_y

    def set_all_fview_plugins(self,plugins):
        pass

    def camera_starting_notification(self,cam_id,
                                     pixel_format=None,
                                     max_width=None,
                                     max_height=None):
        self.target_x = int(max_width / 2)
        self.target_y = int(max_height / 2)
        self.target_x_max = int(max_width)
        self.target_y_max = int(max_height)

    def process_frame(self, cam_id, buf, buf_offset, timestamp, framenumber):
        draw_points = []
        draw_linesegs = []

        now = time.time()
        if (now - self._show_target_time) < 10:
            draw_points.append((self.target_x, self.target_y))

        #hx,hy is the head, bx,by is the body
        target_which = 0
        hx, hy, bx, by = None, None, None, None
        extra = {}
        if self.enabled:
            im = np.array(buf)
            which = self.enable_which
            if which == "head":
                hx, hy, bx, by, extra = self.blob.detect_fly_head(im)
                if hx is not None:
                    draw_points.append((hx, hy))
            elif which == "body":
                bx, by, extra = self.blob.detect_fly_body(im)
                hx, hy = bx, by
                if bx is not None:
                    draw_points.append((bx, by))

            target_which = self._target_map[which]

        if bx is None:
            bx = by = self._no_detect

        if hx is None:
            hx = hy = self._no_detect
        
        contour_area = extra.get('contour_area',-1)
        threshold = extra.get('threshold',-1)

        dt = time.time() - now
        if self._pub is not None:
            self._pub.publish(hx, hy, bx, by, self.target_x, self.target_y, target_which,
                              framenumber, dt, contour_area, threshold)
        else:
            print "dt",dt

        return draw_points, draw_linesegs

