#include <opencv2/opencv.hpp>
#include <opencv2/gpu/gpu.hpp>

#include <iostream>

#include "libcvtempl.h"

#define METH_ANS(_m, _min, _max) \
        (((_m == cv::TM_SQDIFF_NORMED) || (_m == cv::TM_SQDIFF)) ? _min : _max)

static bool _check_value(int method, double val)
{
    if (method == cv::TM_SQDIFF_NORMED)
        return val < 0.9;
    else
        return true;
}

bool is_gpu_supported()
{
    return cv::gpu::getCudaEnabledDeviceCount() > 0;
}

Matcher::Matcher(unsigned int id, int method) :
    m_method(method) 
{
    cv::gpu::setDevice(id);
}

void Matcher::load_template(std::string tmpl)
{
    cv::Mat templ_h = cv::imread(tmpl, CV_LOAD_IMAGE_GRAYSCALE);
    set_template(templ_h.rows, templ_h.cols, templ_h.data);
}

void Matcher::set_template(int height, int width, void *data)
{
    cv::Mat templ_hflip;
    cv::Mat templ_h(height, width, CV_8U, data);
    cv::flip(templ_h, templ_hflip, 0);
    m_tmpl_d.upload(templ_h);
    m_tmpl_dflip.upload(templ_hflip);
}

void Matcher::match_template(int height, int width, void *data, int *loc_x, int *loc_y, double *value, bool *value_ok)
{
    double min_value, max_value;
    cv::Point min_loc, max_loc, loc;

    cv::Mat image_h(height, width, CV_8U, data);

    m_image_d.upload(image_h);
    cv::gpu::matchTemplate(m_image_d, m_tmpl_d, m_result_d, m_method);
    cv::gpu::minMaxLoc(m_result_d, &min_value, &max_value, &min_loc, &max_loc);

    loc = METH_ANS(m_method, min_loc, max_loc);
    *value = METH_ANS(m_method, min_value, max_value);
    *value_ok = _check_value(m_method, *value);

    *loc_x = loc.x;
    *loc_y = loc.y;
}

void Matcher::match_template_both_ways(int height, int width, void *data, int *loc_x, int *loc_y, double *value, bool *value_ok, bool *flipped)
{
    double umin_value, umax_value, fmin_value, fmax_value, uvalue, fvalue;
    cv::Point umin_loc, umax_loc, uloc, fmin_loc, fmax_loc, floc, loc;

    cv::Mat image_h(height, width, CV_8U, data);
    m_image_d.upload(image_h);

    cv::gpu::matchTemplate(m_image_d, m_tmpl_d, m_result_d, m_method);
    cv::gpu::minMaxLoc(m_result_d, &umin_value, &umax_value, &umin_loc, &umax_loc);
    uloc = METH_ANS(m_method, umin_loc, umax_loc);
    uvalue = METH_ANS(m_method, umin_value, umax_value);

    cv::gpu::matchTemplate(m_image_d, m_tmpl_dflip, m_result_d, m_method);
    cv::gpu::minMaxLoc(m_result_d, &fmin_value, &fmax_value, &fmin_loc, &fmax_loc);
    floc = METH_ANS(m_method, fmin_loc, fmax_loc);
    fvalue = METH_ANS(m_method, fmin_value, fmax_value);

    if ((m_method == cv::TM_SQDIFF_NORMED) || (m_method == cv::TM_SQDIFF)) {
        if (uvalue <= fvalue) {
            *flipped = false;
            loc = uloc;
            *value = uvalue;
        } else {
            *flipped = true;
            loc = floc;
            *value = fvalue;
        }
    } else {
        std::cerr << "METHOD NOT SUPPORTED IN FLIPPED CONFIG\n";
        *flipped = false;
        loc = uloc;
        *value = uvalue;
    }

    *value_ok = _check_value(m_method, *value);

    *loc_x = loc.x;
    *loc_y = loc.y;
}

